(define table
  `([main [("apple") ,(lambda () 'apple)]
          [("banana") ,(lambda () 'banana)]
          [(fruit-salad "!") ,(lambda (x)
                                `(,x !))]
          [(fruits2 ".") ,(lambda (x) x)]]
    [fruit-salad [(fruit fruit) ,(lambda (x y)
                                   (list x y))]
                 [(fruit fruit fruit) ,(lambda (x y z)
                                         (list x y z))]]
    [fruit [("orange") ,(lambda () 'orange)]
           [("grape") ,(lambda () 'grape)]]
    [fruits [() ,(lambda () '())]
            [(fruit fruits) ,cons]]
    [fruits2 [(fruits2 fruit) ,(lambda (l v)
                                 (append l (list v)))]
             [() ,(lambda () '())]]))

(define (parse what table stream 
               success-k fail-k)
  (local [(define defn (assq what table))]
    (if defn
        (parse-alts (rest defn) table stream 
                    success-k fail-k)
        (error 'parse "no definition for ~a" what))))

(define (parse-or a-parse b-parse success-k fail-k)
  (local [(define (make-fst-success-k success-k fail-k)
            (lambda (val stream next-fst)
              (success-k val stream
                         (lambda (success-k fail-k)
                           (next-fst (make-fst-success-k success-k fail-k)
                                     (make-fst-fail-k success-k fail-k))))))
          (define (make-fst-fail-k success-k fail-k)
            (lambda ()
              (b-parse success-k fail-k)))]
    (a-parse (make-fst-success-k success-k fail-k)
             (make-fst-fail-k success-k fail-k))))

(define tasks empty)

(define (my-thread t)
  (set! tasks (append tasks (list t))))

(define (my-make-semaphore)
  (box 0))

(define (my-semaphore-post s)
  (set-box! s (add1 (unbox s))))

(define (my-semaphore-wait s k)
  (if (positive? (unbox s))
      (begin
        (set-box! s (sub1 (unbox s)))
        (k))
      (begin
        (my-thread (lambda ()
                     (my-semaphore-wait s k)))
        (my-swap))))

(define (my-swap)
  (local [(define t (first tasks))]
    (set! tasks (rest tasks))
    (t)))

(define (parse-parallel-or a-parse b-parse success-k fail-k)
  (local [(define s (my-make-semaphore))
          (define results empty)
          (define (start-one a-parse)
            (my-thread (lambda ()
                         (a-parse (lambda (val stream next)
                                    (set! results (cons (list val stream next)
                                                        results))
                                    (my-semaphore-post s)
                                    (my-swap))
                                  (lambda ()
                                    (set! results (cons empty
                                                        results))
                                    (my-semaphore-post s)
                                    (my-swap))))))]
    (start-one a-parse)
    (start-one b-parse)
    (local [(define (one-result success-k fail-k)
              (my-semaphore-wait 
               s
               (lambda ()
                 (local [(define result (first results))]
                   (set! results (rest results))
                   (if (= (length result) 3)
                       (success-k (first result)
                                  (second result)
                                  (lambda (success-k fail-k)
                                    (start-one (third result))
                                    (one-result success-k fail-k)))
                       (fail-k))))))]
      (parse-or one-result
                one-result
                success-k
                fail-k))))

(define (parse-alts alts table stream success-k fail-k)
  (if (null? alts)
      (fail-k)
      (parse-parallel-or
       (lambda (success-k fail-k)
         (parse-case (first alts) table stream success-k fail-k))
       (lambda (success-k fail-k)
         (parse-alts (rest alts) table stream success-k fail-k))
       success-k fail-k)))

(define (make-success-k val->val success-k fail-k)
  (local [(define (make-my-success-k success-k fail-k)
            (lambda (val stream next)
              (success-k (val->val val) stream
                         (lambda (success-k fail-k)
                           (next (make-my-success-k success-k fail-k)
                                 fail-k)))))]
    (make-my-success-k success-k fail-k)))

(define (parse-case case table stream success-k fail-k)
  (local [(define seq (first case))
          (define handler (second case))]
    (parse-sequence seq table stream
                    (make-success-k (lambda (vals)
                                      (apply handler vals))
                                    success-k
                                    fail-k)
                    fail-k)))

(define (parse-sequence seq table stream success-k fail-k)
  (cond
    [(empty? seq) (success-k empty stream
                             (lambda (success-k fail-k)
                               (fail-k)))]
    [else 
     ;; This could be abstracted out as "and"
     (parse-item (first seq) table stream
                 (local [(define (make-fst-success-k success-k fail-k)
                           (lambda (fst-vals stream fst-next)
                             (parse-sequence (rest seq) table stream
                                             (make-snd-success-k fst-vals fst-next success-k fail-k)
                                             (lambda ()
                                               (fst-next (make-fst-success-k success-k fail-k)
                                                         fail-k)))))
                         (define (make-snd-success-k fst-vals fst-next success-k fail-k)
                           (lambda (snd-vals stream snd-next)
                             (success-k (append fst-vals snd-vals)
                                        stream
                                        (lambda (success-k fail-k)
                                          (snd-next (make-snd-success-k fst-vals fst-next success-k fail-k)
                                                    (lambda ()
                                                      (fst-next (make-fst-success-k success-k fail-k)
                                                                fail-k)))))))]
                   (make-fst-success-k success-k fail-k))
                 fail-k)]))

(define (parse-item item table stream success-k fail-k)
  (cond
    [(string? item)
     (if (and (cons? stream)
              (equal? item (first stream)))
         (success-k empty (rest stream) 
                    (lambda (success-k fail-k)
                      (fail-k)))
         (fail-k))]
    [else
     (parse item table stream
            (make-success-k list success-k fail-k)
            fail-k)]))

(parse 'fruit-salad table '("orange" "orange" "!")
       (lambda (val stream again) val)
       (lambda () 'nope!))

(parse 'main table '("orange" "grape" "!")
       list
       (lambda () 'nope!))

(parse 'main table '("orange" "grape" "orange" ".")
       list
       (lambda () 'nope!))
